!=============================================================================!
!=============================================================================!
module random_phases
  use bfield_parameters
  implicit none
  save
  double precision, allocatable, dimension(:) :: ca,sa,cp,b,sp,ct,st,s
end module random_phases
!=============================================================================!
!=============================================================================!
module Bgrid
  implicit none
  save
  integer nx,ny,nz
  double precision, allocatable, dimension(:) :: xa,ya,za
  double precision, allocatable, dimension(:,:,:,:) :: Ba
end module Bgrid
!=============================================================================!
!=============================================================================!
subroutine init_turb
  use internal, only : debug;
  implicit none

  call init_phases
  call init_norm

  if (debug>0) call test_turbB
  
end subroutine init_turb
!=============================================================================!
!=============================================================================!
subroutine generate_B(x, Omega)
  use bfield_parameters, only : b_model,B_rms
  implicit none
  double precision Omega(3),x(3)
  
  select case (b_model)
  case (1)
     Omega(1) = B_rms
     Omega(2:3) = 0.d0  
  case (2)
     call generate_B_turb(x, Omega)
  case (3)
     call B_interpolation(x, Omega)
  case default
     call error('unknown value of model_B',0)
  end select

end subroutine generate_B
!=============================================================================!
!=============================================================================!
subroutine generate_B_turb(x, Omega)
  use constants
  use bfield_parameters
  use random_phases
  implicit none
  integer i
  double precision k,l,O_k,d_O(3),Omega(3),x(3)
  complex*16 v(3),ar
  complex*16, parameter :: im = (0.d0,1.d0)
  
  l = (k_max/k_min)**(1.d0/n_k) 
  d_O = 0.d0
  k = k_min
  
  do i=1,n_k

     O_k = O_min * (k/k_min)**(-gamma/2.d0)
     v(1) =  ca(i)*ct(i)*cp(i) - im*s(i)*sa(i)*sp(i)
     v(2) =  ca(i)*ct(i)*sp(i) + im*s(i)*sa(i)*cp(i)
     v(3) = -ca(i)*st(i)
     
     ar =  k * ( st(i)*cp(i)*x(1) + st(i)*sp(i)*x(2) + ct(i)*x(3) ) + b(i) 
     d_O = d_O + O_k * real( v*exp(im*ar) )

     k=l*k

  end do
  
  Omega = d_O    
      
end subroutine generate_B_turb
!=============================================================================!
!=============================================================================!
subroutine init_phases
  use constants
  use random_phases; use bfield_parameters, only : h
  implicit none
  integer i
  double precision r,cc,ss
  double precision ran0

  double precision a,t,p

  allocate (ca(n_k));  allocate (sa(n_k));  allocate (cp(n_k)); 
  allocate (b(n_k));   allocate (sp(n_k));  allocate (ct(n_k));
  allocate (st(n_k));  allocate (s(n_k)); 


  do i=1,n_k
     a = two_pi * ran0()           
     b(i) = two_pi * ran0()
     p = two_pi * ran0()                ! 2-dim: p(i) = 0.d0
     t = acos(1.d0 - 2.d0*ran0())       ! theta distribution p(theta) = sin(theta)/2, 0<theta<pi
     cp(i) = cos(p)
     sp(i) = sin(p)
     ct(i) = cos(t)
     st(i) = sin(t)
     ca(i) = cos(a)
     sa(i) = sin(a)                             ! so that the distribution of radial vectors is uniformly distributed 
                                                ! in spatial angles (dS = sin(t)dtdp) 

     r = 2.d0*ran0()-1.d0                         ! h=-1..1
     if (r>h) s(i)=1

  enddo

end subroutine init_phases
!=============================================================================!
!=============================================================================!
subroutine cs(cc,ss)
  implicit none
  double precision cc,ss,s1,s2,s3
  double precision ran0
  
  do
     s1=2.d0*ran0()-1.d0 
     s2=2.d0*ran0()-1.d0 
     s3=s1*s1+s2*s2
     if(s3<=1.d0) exit
  enddo
  s3=dsqrt(s3)
  cc=s1/s3
  ss=s2/s3      
  
end subroutine cs
!=============================================================================!
!=============================================================================!
subroutine init_norm
  use bfield_parameters
  implicit none
  integer i
  double precision k,N,v

  N = 0.d0
  v = (k_max/k_min)**(1.d0/n_k)
  k = k_min
  do i=1,n_k
     N = N+(k/k_min)**(-gamma)   
     k = v*k
  enddo
  
  O_min = B_rms/sqrt(N)
  O_min = O_min*sqrt(2.d0)                       ! we drop Im(B)

end subroutine init_norm
!=============================================================================!
!=============================================================================!
subroutine init_Bgrid
  use Bgrid
  implicit none
  integer i,j,k

  allocate (xa(nx))
  allocate (ya(ny))
  allocate (za(nz))
  allocate (Ba(nx,ny,nz,3))
  
  open(1,file='Bgrid')
  do i=1,nx
     do j=1,ny
        do k=1,nz
           read(1,*) xa(i),ya(j),za(k),Ba(i,j,k,1),Ba(i,j,k,2),Ba(i,j,k,3)
        end do
     end do
  end do
  close(1)

end subroutine init_Bgrid
!=============================================================================!
!=============================================================================!
subroutine B_interpolation(x,Omega)
  use Bgrid
  implicit none
  integer i,j,l
  integer, save :: i_try,j_try,k_try
  integer, parameter :: n=2,m=3
  double precision x(3),r(n),t,Omega(3)
  
! find nearest array entries of (x,y,z):  
  call hunt(xa,nx,x(1),i_try)
  call hunt(ya,ny,x(2),j_try)
  call hunt(za,nz,x(3),k_try)
  i = min(max(i_try-(m-1)/2,1),nx+1-m)
  j = min(max(j_try-(m-1)/2,1),ny+1-m)

  do l=1,3
  
! perform 2 bilinear interpolations in upper and lower xy plane:
     call bilintp(xa(i),ya(j),Ba(i,j,k_try,l),x(1),x(2),r(1))
     call bilintp(xa(i),ya(j),Ba(i,j,k_try+1,l),x(1),x(2),r(2))

! combine result with linear interpolation in z direction:
     t = (x(3)-za(k_try))/(za(k_try+1)-za(k_try))
     Omega(l)  = (1.-t)*r(1) + t*r(2) 

  end do

end subroutine B_interpolation
!=============================================================================!
!=============================================================================!
subroutine test_turbB
 use  bfield_parameters, only : B_rms,k_min,k_max,gamma
  use constants, only : two_pi
  implicit none
  integer i,j
  double precision L_c,L_max,L_min,z,x(3),B(3),B2(3),alpha

  write(*,*) '!!!!!  test_turbB starts  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'

  
  L_min = two_pi/k_max
  L_max = two_pi/k_min
  alpha=gamma
  
  if (L_min==L_max) then
     L_c = L_max/2.d0
  else
     z = L_min/L_max
     L_c = L_max/2.d0*(alpha-1.d0)/alpha*(1.d0-z**alpha)/(1.d0-z**(alpha-1.d0))
  end if

  write(*,*) "L_max ", L_max, "L_min ", L_min
  write(*,*) "effective coherence length ",L_c, L_max/5.d0, L_max/2.d0
  write(*,*) "Kolmogorov, narrow:",L_max/5.d0, L_max/2.d0
  
  x=0.d0; B2=0.d0; j=0
  do
     x(3)=x(3)+L_min/100.d0
     call generate_B(x,B)
     do i=1,3
        B2(i)=B2(i)+B(i)**2
     end do
     j = j+1
     if (x(3)>100.d0*L_max) exit
  end do
  B2 = sqrt(B2/dble(j))
  write(*,*) ' for isotropy: should be 1:'
  write(*,*) B2/B_rms*sqrt(3.d0)
  write(*,*) ' for correct normalisation: should be 1:'
  write(*,*) sqrt(B2(1)**2+B2(2)**2+B2(3)**2)/B_rms
  
  open(1,file='B'); open(2,file='B2')
  x=0.d0; B2=0.d0
  do
     x(3)=x(3)+L_min/100.d0
     call generate_B(x,B)
     write(1,*) x(3),B
     do i=1,3
        B2(i)=B2(i)+B(i)**2*(L_min/100.d0)
     end do
     write(1,*) x(3),B
     write(2,*) x(3),B2
     if (x(3)>100.d0*L_max) exit
  end do
  close(1); close(2)
  write(*,*) 'B/L_max', sqrt(B2(1)/L_max),sqrt(B2(2)/L_max),sqrt(B2(3)/L_max)

  write(*,*) '!!!!!  test_turbB ends  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
  
end subroutine test_turbB
!=============================================================================!
!=============================================================================!
