!=============================================================================!
!=============================================================================!
module bfield_parameters                    ! values are set in file input_b
   implicit none
   save
   integer :: b_model,n_k
   double precision :: B_rms,gamma,k_min,k_max,h,L_c
   double precision :: O_min                ! calculated in init_norm
end module bfield_parameters
!=============================================================================!
!=============================================================================!
module user_variables
   use constants, only : degree_rad
   implicit none
   save
   logical :: three_dim,psf_homo
   integer :: EBL_model,nmax,en_dist,z_dist
   double precision :: en_max,en_min,en_br,gam1,gam2,m,z_max,th_jet,th_jetx
   
   double precision, parameter ::    &
      ethr=1.d9,                    & ! energy threshold, eV
      ir_rescale=1.d0,              & ! multiplicative rescaling IR background
      a_smp=0.9d0                     ! 1 max. weighted sampling, 0-no sampling

end module user_variables
!=============================================================================!
!=============================================================================!
module user_result
   implicit none
   save
   integer, parameter ::            &  ! for (1+1)-dimensional output        
      n_bin = 81,                   &  ! # of bins in agamma
      n_bin2 = 2,                   &  ! # angular bin
      n_bin3 = 7                       ! # time bins old

   double precision spec(n_bin,0:1),spec_tot(n_bin,0:1)     ! diffuse spectrum
   double precision gam_th(n_bin,n_bin2),gam_th_tot(n_bin,n_bin2) ! spectrum
   double precision spec_t(n_bin,n_bin3),spec_t_tot(n_bin,n_bin3) ! spectrum
   double precision cum_t(n_bin,n_bin3)                     ! cum. spectrum

   
   integer :: n_bint,n_binx,n_biny,n_bind,shiftx    ! for 3-dimensional output
   
   double precision, allocatable, dimension(:,:,:) :: anobs,anobs_tot
   integer, allocatable, dimension(:) :: n_reg
   double precision, allocatable, dimension(:) :: t_bins
   
   integer     :: n_kc                     ! Counting bad particles

   character*4 :: filename = '_mpi'         ! name in output

 end module user_result
!=============================================================================!
!=============================================================================!
subroutine user_main(myid,nmax)
!-----------------------------------------------------------------------------!
!           calls: plots,initial_particle,cascade                             !
!-----------------------------------------------------------------------------!
   use user_result
   implicit none
   integer myid,nl,icq,nmax
   double precision z,e0,weight

   if (myid==0) then
      call plots
      write(*,*)'start with nmax = ',nmax
   end if
   spec=0.d0
   gam_th=0.d0
   spec_t=0.d0
   cum_t=0.d0
   anobs=0.d0
   n_reg = 0
   
   do nl=1,nmax
      call initial_particle(e0,weight,z)           ! initial energy & redshift
      icq = 0                                      ! (0 - gamma, +-1 - e+-)
      call cascade(icq,e0,weight,z)                ! starts EM-cascade
      !call register(e0,x,x,weight,x,icq)
      if (myid.eq.0.and.mod(nl*100,nmax).eq.0) write(*,*)'nl=',nl
   enddo

 end subroutine user_main
!=============================================================================!
!          sample initial photon/electron energy and redshift                 !
!=============================================================================!
subroutine initial_particle(e0,weight,z)
!-----------------------------------------------------------------------------
! output:
!        e0       - particle energy;
!        weight   - initial particle weight;
!        z        - initial redshift;
!-----------------------------------------------------------------------------
  use user_variables
  implicit none
  double precision e0,weight,z,r,r_max 
  double precision psran,w_spec,r_z,z_r

  select case (en_dist)
  case (1)                                            ! fixed energy
     e0 = en_br
     weight=(e0/en_br)**(-gam1+1.d0)*log(en_max/en_min)
  case (2)                                   ! broken power-law dN/dE ~ E0^gam
     e0=en_min*(en_max/en_min)**psran()              ! energy: uniform in ln E
     if (e0<en_br) then
        weight=(e0/en_br)**(-gam1+1.d0)*log(en_max/en_min) 
     else
        weight=(e0/en_br)**(-gam2+1.d0)*log(en_max/en_min) 
     end if
  case (3)                                           ! user-defined dN/dE 
     e0=en_min*(en_max/en_min)**psran()              ! energy: uniform in ln E
     weight = w_spec(e0)
  case default
     write(*,*) 'en_dist = ',en_dist
     call error('wrong value of en_dist',0)
  end select  

  select case (z_dist)
  case (1)                                            ! fixed redshift
     z = z_max
  case (2)                                                    ! ~(1+z)^m
     r_max = r_z(z_max)
     r = r_max * psran()                        ! uniform in comoving distance
     z = z_r(r)
     weight = weight * (1.d0+z)**m
  case default
     write(*,*) 'z_dist = ',z_dist
     call error('wrong value of z_dist',0)
  end select

end subroutine initial_particle
!=============================================================================!
!=============================================================================!
!       user-defined generation spectrum  f=E*dN/dE at the source             !
!=============================================================================!
double precision function w_spec(E)
!-----------------------------------------------------------------------------!
!        input    :  E       -  injection energy                              !
!        output   :  w_spec  -  weight                                        !
!-----------------------------------------------------------------------------!
   implicit none
   double precision E,f
   double precision, parameter :: Ecut = 1.d13,alpha=2.2d0

   f = E**(1.d0-alpha) * exp(-E/Ecut)
   w_spec = f
  
end function w_spec
!=============================================================================!
!        Additional weighting according to a jet distribution                 !
!=============================================================================!
subroutine jet_distribution(the_s,w_jet)
!-----------------------------------------------------------------------------!
!        input    :  the_s    -  angle from jet center
!        output   :  w_jet    -  weigth given jet distribution
!-----------------------------------------------------------------------------!
   use user_variables, only : th_jet
   implicit none
   double precision the_s,w_jet
! radial gaussian distribution with weight 1 at the center and
   w_jet=exp(-the_s**2/(2.d0*th_jet**2))  ! gaussian distribution
   !w_jet=1                       ! uniform
end subroutine jet_distribution
!=============================================================================!
!=============================================================================!
subroutine register(e0,thex,they,weight,dt,icq)
   use user_variables
   use user_result
   use constants, only : pi,degree_rad
   implicit none
   integer icq,i,j,k
   double precision e0,weight,thex,they,dt
   double precision thereg_en,theta,theta95
  
   ! diffuse energy spectrum:
   i=min(n_bin,int(log(e0/ethr)/log(en_max/ethr)*(n_bin-1))+1)
   i=max(i,1)
   spec(i,abs(icq))=spec(i,abs(icq))+weight*e0/log(en_max/ethr)*(n_bin-1)

   if (icq.ne.0) return                                ! forget electrons

   if (three_dim) call psf_spread(e0,thex,they,weight,dt)

   theta = sqrt( thex**2+they**2 )
   theta95 = thereg_en(e0)
   if (theta<theta95) then
      j=1
   else
      j=2
   end if
   gam_th(i,j)=gam_th(i,j)+weight*e0/log(en_max/ethr)*(n_bin-1)
   if (j.eq.1) then                                ! only gamma's inside PSF
      if (theta.eq.0.d0.or.dt.eq.0) then
         k=1
      else
         k = int(log10(dt))+1
      end if
      if (k<1) k=1
      if (k>n_bin3) k=n_bin3
      spec_t(i,k)=spec_t(i,k)+weight*e0/log(en_max/ethr)*(n_bin-1)
   end if

end subroutine register
!=============================================================================!
!        distributing detected particle according to a PSF function           !
!=============================================================================!
subroutine psf_spread(e0,thex,they,weight,dt)
   use user_result; use user_variables, only : psf_homo
   use constants, only : two_pi,rad_degree
   implicit none
   integer i,j,k
   double precision e0,thex,they,weight,dt
   double precision rx,ry,r,sig,thereg_en

   if (dt>t_bins(n_bint-1)) return                 ! outside time range, forget

   if (dt<t_bins(n_bint)) then                      ! non-interacting particles
      k = n_bint                                    ! into last time bin
   else                                             ! find k or others
      k=1
      do
         if (dt<t_bins(k)) exit
         k=k+1
      end do
   end if
   n_reg(k) = n_reg(k) + 1


   sig=thereg_en(e0)
   do i=1,n_binx
      rx=(i-n_binx/2.d0)/dble(n_bind) + shiftx
      do j=1,n_biny
         ry=(j-n_biny/2.d0)/dble(n_bind)
         r=sqrt((rx-thex)**2 + (ry-they)**2)
         if (psf_homo) then            ! homogeneous distribution of PSF
            if (r.lt.sig) anobs(k,i,j) = anobs(k,i,j) + weight*e0/sig**2
         else                          ! Gaussian 
            anobs(k,i,j) = anobs(k,i,j) + weight*e0*exp(-2.d0*r**2/sig**2)
         end if
      end do
   end do
   
end subroutine psf_spread
!=============================================================================!
!        theta_PSF/degree (95%) from Fermi and HESS                           !
!=============================================================================!
double precision function thereg_en(en)
   implicit none
   double precision en

   if (en.lt.1.d6) then                            ! avoid theta>360 degrees 
      thereg_en = 359.d0
      return
   end if

   if (en.le.3e11) then                                             ! Fermi
      thereg_en = 10.d0**7.12d0*en**(-0.766d0)             
      if (en>1d9) thereg_en = thereg_en + 0.2d0*exp(-1.d10/en)
   else                                                             ! HESS
      thereg_en = 0.11d0                                      
   endif
end function thereg_en
!=============================================================================!
!=============================================================================!
subroutine user_output(n_max,n_proc) 
   use user_variables
   use user_result
   implicit none
   integer n_max,n_proc,i,j
   character*5 str

   spec_tot = spec_tot/dble(n_proc*n_max)
   gam_th_tot = gam_th_tot/dble(n_proc*n_max)
   spec_t_tot = spec_t_tot/dble(n_proc*n_max)
   anobs_tot = anobs_tot/dble(n_proc*n_max)
   
   cum_t(:,1) = spec_t_tot(:,1)
   do i=2,n_bin3
      cum_t(:,i) = cum_t(:,i-1)+spec_t_tot(:,i)
   end do

   open(unit=31,file='Output/spec_diff'//filename) ! energy spectrum g,e
   open(unit=32,file='Output/spec_95'//filename)   ! energy spectrum g in PSF
   open(unit=33,file='Output/spec_t'//filename)    ! energy-time spectra g in PSF
   open(unit=34,file='Output/spec_c'//filename)    ! energy-time spectra g in PSF
   do i=1,n_bin-1
      write(31,*) real(ethr*(en_max/ethr)**((i-.5d0)/(n_bin-1))), &
                  real(spec_tot(i,0)),real(spec_tot(i,1))
      write(32,*) real(ethr*(en_max/ethr)**((i-.5d0)/(n_bin-1))), &
                  real(gam_th_tot(i,1)),real(gam_th_tot(i,2))
      write(33,*) real(ethr*(en_max/ethr)**((i-.5d0)/(n_bin-1))), &
                  real(spec_t_tot(i,:))
      write(34,*) real(ethr*(en_max/ethr)**((i-.5d0)/(n_bin-1))), &
                  real(cum_t(i,:))
   end do
   close(31)
   close(32)
   close(33)
   close(34)

   do j=1,n_bint
      write (str,'(I1.1)') j
      open(unit=35,file='Output/AngRes/angle_matrix'//filename//str)
      write(35,*) th_jetx, n_binx, n_biny, shiftx, n_bind
      do i=1,n_binx
         write(35,*) real(anobs_tot(j,i,:))
      end do
      close(35)
   end do
   write(*,*) "jet off by", th_jetx, "degrees"

end subroutine user_output
!=============================================================================!
!=============================================================================!
subroutine plots
   implicit none
   integer icq
   double precision E,x(3),z0,rate(2),psf95
   double precision rate_EBL_tab,thereg_en

   x(1) = 125.*3.086d24
   x(2) = 808.*3.086d24
   x(3) = 1500.*3.086d24

   z0 = 0.d0
   icq = 0 

   E = 1.d8
   open(41,file='Output/lint')
   open(42,file='Output/tau')
   open(43,file='Output/psf95')
   do
      E=E*1.1d0
      rate(1) = rate_EBL_tab(E,z0,icq)
      rate(2) = rate_EBL_tab(E,z0,1)
      write(41,*) real(E),rate(1)*3.086d24,rate(2)*3.086d24
      psf95 = thereg_en(E)
      write(43,*) real(E),psf95
      if (E>1d21) exit
   end do
   close(41)
   close(42)
   close(43)
end subroutine plots
!=============================================================================!
!=============================================================================!
